#if 1
//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

///
///@file
///This file implements the "test" builtin command.  It has 2 names:  "test" and "[". 
///
/// The functions in this file named parse* are acutally "parse and execute"*.  Sorry for the name confusion.


#include "TestExpression.h"
#include <cxxtls/scriptthread.h>
#include <cxxtls/lexical_cast.h>
#include <sys/stat.h>
#include <vector>
#include <string>
#include <windows.h>

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

namespace BrashTestExpression
{

static string eqToken("=");
static string neToken("!=");
static string intEqToken("-eq");
static string intGtToken("-gt");
static string intGeToken("-ge");
static string intLtToken("-lt");
static string intLeToken("-le");
static string intNeToken("-ne");
static string xToken("-x");
static string dashO("-o");
static string dashA("-a");
static string bangToken("!");
static string closeSquare("]");
static string regexEq("=~");
static string dashF("-f");
static string dashE("-e"); // is regular, the file exists but is not a directory
static string dashR("-r");
static string dashW("-w");
static string dashD("-d");
static string dashX("-x");
static string dashS("-s");  // exists and is not empty
static string dashT("-t");  // is terminal
static string dashN("-n");  // string length not zero
static string dashZ("-z");  // string length IS zero

#if 1

bool FileOpExpression::fileNumberIsTerminal(ThreadContext *context, string const &fileNumberString)
{
    int fileno = -1;

    try
    {
        fileno = cxxtls::lexical_cast<int>(fileNumberString);
    }
    catch(...)
    {
        throw BadFileNumberException("-t option parameter", fileNumberString);
    }

    cxxtls::ScriptThread::ManagedFileHandles::iterator ptr = context->fileHandles_->find(fileno);

    if(ptr != context->fileHandles_->end())
    {
        if(ptr->second.error_)
           return false; // file has to be open

        cxxtls::ScriptThread::ManagedFileHandles::FileHandleAttributes::handle_t osHandle = ptr->second.getOSHandle();

        #ifdef _MSC_VER
            DWORD consoleMode;
    
    
            if(GetConsoleMode(osHandle, &consoleMode))
            {
               return true;
            }

        #else
            return isatty(osHandle);
        #endif

    }


    return false;
}



TestExpression *parseBinop(vector<string>::const_iterator &nextToken, vector<string>::const_iterator const &lastToken)
{
   // assumes that the leading square bracket has been already discarded from the stream.  We are parsing
   //   one of these syntaxes:
   //
   //   [ ! -f fred ]
   //       ^
   //
   //   [ a > b ]
   //     ^
   //
   //   [ aString ]
   //     ^
   //


    if(nextToken == lastToken)
    {

       throw TestCommandSyntaxMissingExpression();

       return 0;
    }


   bool inverted = false;


   if(nextToken != lastToken && *nextToken == bangToken)
   {
       inverted = true;

       ++nextToken;
   }


   string const &firstToken = *nextToken;

   if(firstToken == "-t")
   {
      // this is a test to see if a numeric token refers to an open file handle

      ++nextToken;

      if(nextToken == lastToken)
      {
        throw TestCommandSyntaxMissingExpression("parameter for", firstToken);
      }

      auto_ptr<FileOpExpression> e(new FileOpExpression(FILEOP_ISTERM, *nextToken++));

      e->setInverted(inverted);

      return e.release();


   }
   else
   if(!firstToken.empty())
   {
       // we might have a fileop



       FILEOP op = FILEOP_UNSPEC;

       if (firstToken == dashE)
           op = FILEOP_REGULAR;
       else
       if (firstToken == dashF)
           op = FILEOP_EXISTS;
       else
       if (firstToken == dashR)
           op = FILEOP_READABLE;
       else
       if (firstToken == dashW)
           op = FILEOP_WRITEABLE;
       else
       if (firstToken == dashD)
           op = FILEOP_ISDIR;
       else
       if (firstToken == dashX)
           op = FILEOP_EXE;
       else
       if (firstToken == dashS)
           op = FILEOP_NOTEMPTY;
       else
       if (firstToken == dashN)
           op = FILEOP_STR_NEMPTY;
       else
       if (firstToken == dashZ)
           op = FILEOP_STR_EMPTY;
       
       if(op != FILEOP_UNSPEC)
       {
          vector<string>::const_iterator after = nextToken;  

          ++after;

          if(after == lastToken)
          {
            op = FILEOP_UNSPEC; // nothing comes after the fileop so there is a syntax error

            throw TestCommandSyntaxMissingFileName("afterToken", firstToken);

          }
          else
          {
             // we have a first token that could be a fileop, but let us check to see if
             // the token after is one of the comparison operators

             string const &afterToken = *after;

             if(   afterToken == eqToken
                || afterToken == neToken
                || afterToken == intEqToken
                || afterToken == intNeToken
                || afterToken == intGtToken
                || afterToken == intLtToken
                || afterToken == intGeToken
                || afterToken == intLeToken
                || afterToken == regexEq
               )
             {
                op = FILEOP_UNSPEC;  // we have something like this:  [ -f -lt -g ]
             }

          }
       }


       if(op != FILEOP_UNSPEC)
       {
          // ok, we know know we have something like this:
          // [ -f fred ]
          //   ^

          ++nextToken;  // we've already interpreted it

          auto_ptr<FileOpExpression> e(new FileOpExpression(op, *nextToken++));

          e->setInverted(inverted);

          return e.release();


       }

      
   }  

   // if we get here, we know that the first word in the stream is not a file operation
   //
   //  [ fred = tom ]
   //    ^
   //
   //  firstToken = *nextToken at this point


   ++nextToken;

   if(nextToken == lastToken || *nextToken == closeSquare)
   {
      // we have hit the end of the stream with only the one token



      auto_ptr<StringExpression> e(new StringExpression(firstToken));

      e->setInverted(inverted);

      return e.release();

   }


   string const &secondToken = *nextToken++;

   if(nextToken == lastToken || *nextToken == closeSquare)
   {
      throw TestCommandSyntaxBinaryExpressionMissing2ndOperand("firstOperand", firstToken, "operator", secondToken);

      return 0; 
   }

   // at this point, the 3'rd token, which is the second operand is in *nextToken

   auto_ptr<TestExpression> e(0);

   if(secondToken == eqToken)
   {
      e.reset(new StringEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == intEqToken)
   {
      e.reset(new IntStringEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == neToken)
   {
      e.reset(new StringNotEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == intNeToken)
   {
      e.reset(new IntStringNotEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == intGtToken)
   {
      e.reset(new IntStringGreaterExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == intLtToken)
   {
      e.reset(new IntStringLessExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == intLeToken)
   {
      e.reset(new IntStringLessEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == intGeToken)
   {
      e.reset(new IntStringGreaterEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == regexEq)
   {
      e.reset(new RegexEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   {
       throw TestCommandSyntaxInvalidOperand("expected comparison operator but found", secondToken);

       return 0;
   }

   e->setInverted(inverted);

   return e.release();


}



TestExpression *parseFactor(vector<string>::const_iterator &nextToken, vector<string>::const_iterator const &lastToken)
{
   // assumes that the leading square bracket has been already discarded from the stream.  We are parsing
   // this syntax:
   //
   //   [ -f fred -a -d bill ]
   //     ^

   auto_ptr<TestExpression> rv(0);


   if(nextToken == lastToken)
   {
      throw TestCommandSyntaxMissingExpectedToken("missing token", "]");

      return 0;
   }

   while(nextToken != lastToken)
   {
       if( closeSquare == *nextToken)
       {
           break;
       }


       auto_ptr<TestExpression> exp(parseBinop(nextToken, lastToken));

       if(!exp.get())
       {
          // syntax error, message will already have been printed

          return 0;  // rv will be auto destructed
       }

       if(rv.get())
       {
          // we already had a left hand side and now we are adding the right

          auto_ptr<TestExpression> newbie(new AndExpression(rv.release(), exp.release()));

          // rv is now empty, and newbie has the old rv as it's left hand side, and a newly
          // gathered term expression as it's right hand side

          rv.reset(  newbie.release() );  // newbie is empty, and now rv is populated
       }
       else
       {
          // this the first term in the expression

          rv.reset(exp.release());  // exp is now empty and rv has it's term

       }

       if(nextToken == lastToken || *nextToken != dashA)
           break;

       ++nextToken;  // eat the -a


   }

   if(!rv.get())
   {
      throw TestCommandSyntaxMissingExpectedToken("missing whole expression", "");
   }


   return rv.release();
}


TestExpression *parseTerm(vector<string>::const_iterator &nextToken, vector<string>::const_iterator const &lastToken)
{
   // assumes that the leading square bracket has been already discarded from the stream.  We are parsing
   // this syntax:
   //
   //   [ -f fred -a -d bill ]
   //     ^

   auto_ptr<TestExpression> rv(0);


   if(nextToken == lastToken)
   {
      throw TestCommandSyntaxMissingExpectedToken("missing token", "]");

      return 0;
   }


   if(nextToken != lastToken && *nextToken != closeSquare)
   {

       auto_ptr<TestExpression> exp(parseFactor(nextToken, lastToken));

       if(!exp.get())
       {
          // syntax error, message will already have been printed

          return 0;  // rv will be auto destructed
       }

       if(rv.get())
       {
          // we already had a left hand side and now we are adding the right

          auto_ptr<TestExpression> newbie(new AndExpression(rv.release(), exp.release()));

          // rv is now empty, and newbie has the old rv as it's left hand side, and a newly
          // gathered term expression as it's right hand side

          rv.reset(  newbie.release() );  // newbie is empty, and now rv is populated
       }
       else
       {
          // this the first term in the expression

          rv.reset(exp.release());  // exp is now empty and rv has it's term

       }


   }

   if(!rv.get())
   {
      throw TestCommandSyntaxMissingExpectedToken("missing whole expression", "");
   }


   return rv.release();
}




TestExpression *parseTestExpression(vector<string>::const_iterator &nextToken, vector<string>::const_iterator const &lastToken)
{
   // assumes that the leading square bracket has been already discarded from the stream.  We are parsing
   // this syntax:
   //
   //   [ -f fred -o -d bill ]
   //     ^

   auto_ptr<TestExpression> rv(0);

   if(nextToken == lastToken)
   {
      throw TestCommandSyntaxMissingExpectedToken("expected token", "]");


      return 0;
   }

   while(nextToken != lastToken)
   {
       if( closeSquare == *nextToken)
       {
           break;
       }

       auto_ptr<TestExpression> exp(parseTerm(nextToken, lastToken));

       if(!exp.get())
       {
          // syntax error, message will already have been printed

          return 0;  // rv will be auto destructed
       }

       if(rv.get())
       {
          // we already had a left hand side and now we are adding the right

          auto_ptr<TestExpression> newbie(new OrExpression(rv.release(), exp.release()));

          // rv is now empty, and newbie has the old rv as it's left hand side, and a newly
          // gathered term expression as it's right hand side

          rv.reset(  newbie.release() );  // newbie is empty, and now rv is populated
       }
       else
       {
          // this the first term in the expression

          rv.reset(exp.release());  // exp is now empty and rv has it's term

       }

       if(nextToken == lastToken || *nextToken != dashO)
          break;  // only keep going if we  currently see -o in the stream

       ++nextToken;  // eat the -o


   }

   if(!rv.get())
   {
      throw TestCommandSyntaxMissingExpectedToken("missing whole expression", "");
   }

   if(nextToken != lastToken && *nextToken == closeSquare)
      ++nextToken; // eat the trailing ] in the expresssion

   return rv.release();
}

#endif

} // namespace BrashTestExpression

#endif
